#include "RGBLED.h"

byte colorRedMax;
byte colorGreenMax;
byte colorBlueMax;

byte colorRedCurrent;
byte colorGreenCurrent;
byte colorBlueCurrent;

byte colorRedStep;
byte colorGreenStep;
byte colorBlueStep;



// Constructor of Class RGBLED   (function name = class name)
// Parameters are pin numbers connected to single LEDs of RGB LED
// note: implementation of this function is hardware dependent (=ESP32) 
RGBLED::RGBLED (byte redPin, byte greenPin, byte bluePin)
{
	// store pin numbers in object variables
	
	this->redPin   = redPin;
	this->greenPin = greenPin;
	this->bluePin  = bluePin;

    // PWM setup
	
	// red LED
	ledcSetup(0, 5000, 8);       // Channel number, frequency, resolution in bit
	ledcAttachPin(redPin, 0);    // pin number, channel number

	// green LED
	ledcSetup(1, 5000, 8);       // Channel number, frequency, resolution in bit
	ledcAttachPin(greenPin, 1);  // pin number, channel number
	
	// blue LED
	ledcSetup(2, 5000, 8);       // Channel number, frequency, resolution in bit
	ledcAttachPin(bluePin, 2);   // pin number, channel number

}

// private function

// set PWM for all 3 LEDs
// note: this function is hardware dependent (=ESP32) 
void pwmRGB(byte Red, byte Green, byte Blue)
{
	// set PWM duty cycle (here 0..255)
	ledcWrite(0, Red );     // Channel number, duty cycle
	ledcWrite(1, Green );   // Channel number, duty cycle
	ledcWrite(2, Blue );    // Channel number, duty cycle
	
}


// public functions  (classname::functionname)
// note: independent from hardware

// set color for RGB LED, color values 0..255, store the color as default color
void RGBLED::setRGB(byte colorRed, byte colorGreen, byte colorBlue)
{
	// store the color values as a "default color"
	colorRedMax = colorRed;
	colorGreenMax = colorGreen;
	colorBlueMax = colorBlue;
    
	// current color for dimming
	colorRedCurrent = colorRed;
	colorGreenCurrent = colorGreen;
	colorBlueCurrent = colorBlue;

	// color steps for dimming
	colorRedStep = colorRed >> 4;
	colorGreenStep = colorGreen >> 4;
	colorBlueStep = colorBlue >> 4;
	
	// set the color with PWM
	pwmRGB(colorRed, colorGreen, colorBlue);
}


// step forward in dimming cycle:
// dim from default color to 0, then set the default color again
void RGBLED::dimRGB()
{
	if ((colorRedCurrent < colorRedStep) || (colorGreenCurrent < colorGreenStep) || (colorBlueCurrent < colorBlueStep))
	{
		// set default color again
		colorRedCurrent = colorRedMax;
		colorGreenCurrent = colorGreenMax;
		colorBlueCurrent = colorBlueMax;
	}
	else
	{
		// dim current color
		colorRedCurrent = colorRedCurrent - colorRedStep;
		colorGreenCurrent = colorGreenCurrent - colorGreenStep;
		colorBlueCurrent = colorBlueCurrent - colorBlueStep;
	}

	// set current color with PWM
	pwmRGB(colorRedCurrent, colorGreenCurrent, colorBlueCurrent);
}



// Switch LEDs of RGBLED to certain color
// LEDColor must be LED_GREEN, LED_RED, LED_YELLOW, LED_OFF or LED_ON
void RGBLED::SwitchRGBLED(byte LEDColor)
{
	switch(LEDColor)
	{
		case LED_GREEN:
		pwmRGB(0, 255, 0);    // set green with PWM
		break;
		
		case LED_RED:
		pwmRGB(255, 0, 0);    // set red with PWM
		break;
		
		case LED_YELLOW:
		pwmRGB(200, 200, 0);  // set yellow with PWM
		break;
		
		case LED_OFF:
		pwmRGB(0, 0, 0);      // switch off LED with PWM
		break;
		
		case LED_ON:
		pwmRGB(colorRedMax, colorGreenMax, colorBlueMax);   //  switch to default color with PWM
		break;
	}
};