/*
IOT Sensor Board AllThingsTalk Demo with configuration webpage
for ESP32 DevKitC
Elektor / Jens Nickel
My Journey in the IoT
*/


#include <WiFi.h>

#include <Preferences.h>   // this library is used to get access to Non-volatile storage (NVS) of ESP32
// see https://github.com/espressif/arduino-esp32/blob/master/libraries/Preferences/examples/StartCounter/StartCounter.ino






// +++++++++++++++++++ Start of Webserver Library +++++++++++++++++++

// as ESP32 Arduino Version

WiFiClient myclient;
WiFiServer server(80);


void Webserver_Start()
{
  server.begin();     // Start TCP/IP-Server on ESP32
}



String WebRequestHostAddress;     // global variable used to store Server IP-Address of HTTP-Request


//  Call this function regularly to look for client requests
//  template see https://github.com/espressif/arduino-esp32/blob/master/libraries/WiFi/examples/SimpleWiFiServer/SimpleWiFiServer.ino
//  returns empty string if no request from any client
//  returns GET Parameter: everything after the "/?" if ADDRESS/?xxxx was entered by the user in the webbrowser
//  returns "-" if ADDRESS but no GET Parameter was entered by the user in the webbrowser
//  remark: client connection stays open after return
String Webserver_GetRequestGETParameter()
{
  String GETParameter = "";
  
  myclient = server.available();   // listen for incoming clients

  //Serial.print(".");
  
  if (myclient) {                            // if you get a client,
    Serial.println("New Client.");           // print a message out the serial port
    String currentLine = "";                 // make a String to hold incoming data from the client
    
    while (myclient.connected()) {           // loop while the client's connected
      
      if (myclient.available()) {            // if there's bytes to read from the client,
        
        char c = myclient.read();            // read a byte, then
        Serial.write(c);                     // print it out the serial monitor

        if (c == '\n') {                     // if the byte is a newline character

          // if the current line is blank, you got two newline characters in a row.
          // that's the end of the client HTTP request
          if (currentLine.length() == 0) {
            
            if (GETParameter == "") {GETParameter = "-";};    // if no "GET /?" was found so far in the request bytes, return "-"
            
            // break out of the while loop:
            break;
        
          } else {    // if you got a newline, then clear currentLine:
            currentLine = "";
          }
          
        } else if (c != '\r') {  // if you got anything else but a carriage return character,
          currentLine += c;      // add it to the end of the currentLine
        }

        if (c=='\r' && currentLine.startsWith("GET /?")) 
        // we see a "GET /?" in the HTTP data of the client request
        // user entered ADDRESS/?xxxx in webbrowser, xxxx = GET Parameter
        {
          
          GETParameter = currentLine.substring(currentLine.indexOf('?') + 1, currentLine.indexOf(' ', 6));    // extract everything behind the ? and before a space
                    
        }


        if (c=='\r' && currentLine.startsWith("Host:")) 
        // we see a "Host:" in the HTTP data of the client request
        // user entered ADDRESS or ADDRESS/... in webbrowser, ADDRESS = Server IP-Address of HTTP-Request
        {
          int IndexOfBlank = currentLine.indexOf(' ');
          WebRequestHostAddress = currentLine.substring(IndexOfBlank + 1, currentLine.length());    // extract everything behind the space character and store Server IP-Address of HTTP-Request
          
        }
        
      }
      
    }
    
  }

  return GETParameter;
}



// Send HTML page to client, as HTTP response
// client connection must be open (call Webserver_GetRequestGETParameter() first)
void Webserver_SendHTMLPage(String HTMLPage)
{
   String httpResponse = "";

   // begin with HTTP response header
   httpResponse += "HTTP/1.1 200 OK\r\n";
   httpResponse += "Content-type:text/html\r\n\r\n";

   // then the HTML page
   httpResponse += HTMLPage;

   // The HTTP response ends with a blank line:
   httpResponse += "\r\n";
    
   // send it out to TCP/IP client = webbrowser 
   myclient.println(httpResponse);

   // close the connection
   myclient.stop();
    
   Serial.println("Client Disconnected.");
   
};



// +++++++++++++++++++ End of Webserver library +++++++++++++++++++++




// +++++++++++++++++++ Start of WiFi Library ++++++++++++++++++++++++


// Connect to router network and return 1 (success) or -1 (no success)
int WiFi_RouterNetworkConnect(char* txtSSID, char* txtPassword)
{
  int success = 1;
  
  // connect to WiFi network
  // see https://www.arduino.cc/en/Reference/WiFiBegin
  
  WiFi.begin(txtSSID, txtPassword);
  
  // we wait until connection is established
  // or 10 seconds are gone
  
  int WiFiConnectTimeOut = 0;
  while ((WiFi.status() != WL_CONNECTED) && (WiFiConnectTimeOut < 10))
  {
     // try it 10 times a second
    delay(1000);
    WiFiConnectTimeOut++;
  }

  // not connected
  if (WiFi.status() != WL_CONNECTED)
  {
    success = -1;
  }

  // print out local address of ESP32 in Router network (LAN)
  Serial.println(WiFi.localIP());

  return success;
}



// Disconnect from router network and return 1 (success) or -1 (no success)
int WiFi_RouterNetworkDisconnect()
{
  int success = -1;
  
  WiFi.disconnect();
  

  int WiFiConnectTimeOut = 0;
  while ((WiFi.status() == WL_CONNECTED) && (WiFiConnectTimeOut < 10))
  {
    // try it 10 times a second
    delay(1000);
    WiFiConnectTimeOut++;
  }

  // not connected
  if (WiFi.status() != WL_CONNECTED)
  {
    success = 1;
  }
  
  Serial.println("Disconnected.");

  return success;
}


// Initialize Soft Access Point with ESP32
// ESP32 establishes its own WiFi network, one can choose the SSID
int WiFi_AccessPointStart(char* AccessPointNetworkSSID)
{ 
  WiFi.softAP(AccessPointNetworkSSID);

  // printout the ESP32 IP-Address in own network, per default it is "192.168.4.1".
  Serial.println(WiFi.softAPIP());

  return 1;
}


// Put ESP32 in both modes in parallel: Soft Access Point and station in router network (LAN)
void WiFi_SetBothModesNetworkStationAndAccessPoint()
{
  WiFi.mode(WIFI_AP_STA);
}


// Get IP-Address of ESP32 in Router network (LAN), in String-format
String WiFi_GetOwnIPAddressInRouterNetwork()
{
  return WiFi.localIP().toString();
}



// +++++++++++++++++++ End of WiFi Library +++++++++++++++++++










Preferences preferences;   // we must generate this object of the preference library

#define PIN_LED_BOARD  12
#define PIN_ACTOR      12   // the digital output we want to control

#define PIN_LED_GREEN  27   // pin numbers connected to single LEDs of RGB LED
#define PIN_LED_RED    26
#define PIN_LED_BLUE   25

#define PIN_SENSOR     36   // the analog input for sensor

#define LED_ON    128
#define LED_OFF   0

#define ACTOR_ON    1
#define ACTOR_OFF   0



// Switch Actor Output
// ActorState must be ACTOR_OFF or ACTOR_ON
void SwitchActor(byte ActorState)
{
    if (ActorState == ACTOR_OFF)
    {
      digitalWrite(PIN_ACTOR, LOW);
    }
    else
    {      
      digitalWrite(PIN_ACTOR, HIGH);
    }
}


// Switch LED on the board
// LEDState must be LED_OFF or LED_ON
void SwitchBoardLED(byte LEDState)
{
    if (LEDState == LED_OFF)
    {
      digitalWrite(PIN_LED_BOARD, LOW);
    }
    else
    {      
      digitalWrite(PIN_LED_BOARD, HIGH);
    }
}







// this section handles configuration values which can be configured via webpage form in a webbrowser

// 8 configuration values max
String ConfigName[8];     // name of the configuration value
String ConfigValue[8];    // the value itself (String)
int    ConfigStatus[8];   // status of the value    0 = not set    1 = valid   -1 = not valid
int    ConfigType[8];     // type of the value    0 = not set    1 = String (textbox)   2 = Byte (slider)


// Initalize the values 
void InitializeConfigValues()
{
  for (int count = 0; count < 8; count++)
  {
    ConfigName[count] = "";
    ConfigValue[count] = "";
    ConfigStatus[count] = 0;
    ConfigType[count] = 0;
  }
}


// Build a HTML page with a form which shows textboxes to enter the values
// returns the HTML code of the page
String EncodeFormHTMLFromValues(String TitleOfForm, int CountOfConfigValues)
{
   // Head of the HTML page
   String HTMLPage = "<!DOCTYPE html><html><body><h2>" + TitleOfForm + "</h2><form><table>";

   // for each configuration value
   for (int c = 0; c < CountOfConfigValues; c++)
   {
      // set background color by the status of the configuration value
      String StyleHTML = "";
      if (ConfigStatus[c] == 0) { StyleHTML = " Style =\"background-color: #FFE4B5;\" " ;};   // yellow
      if (ConfigStatus[c] == 1) { StyleHTML = " Style =\"background-color: #98FB98;\" " ;};   // green
      if (ConfigStatus[c] == -1) { StyleHTML = " Style =\"background-color: #FA8072;\" " ;};  // red

      String TableRowHTML = "";
  
      if (ConfigType[c] == 1)    //  string value, entered in text box
      {
        // build the HTML code for a table row with configuration value name and the value itself inside a textbox   
        TableRowHTML = "<tr><th>" + ConfigName[c] + "</th><th><input name=\"" + ConfigName[c] + "\" value=\"" + ConfigValue[c] + "\" " + StyleHTML + " /></th></tr>";
      }
  
      if (ConfigType[c] == 2)
      {
        // build the HTML code for a table row with configuration value name and the byte value represented by a slider 
        TableRowHTML = "<tr><th>" + ConfigName[c] + "</th><th><input type=\"range\" min=\"1\"  max=\"255\" name=\"" + ConfigName[c] + "\" value=\"" + ConfigValue[c] + "\" " + StyleHTML + " /></th></tr>";
      }
      
      // add the table row HTML code to the page
      HTMLPage += TableRowHTML;
   }

   // add the submit button
   HTMLPage += "</table><br/><input type=\"submit\" value=\"Submit\" />";

   // footer of the webpage
   HTMLPage += "</form></body></html>";
   
   return HTMLPage;
}





// Decodes a GET parameter (expression after ? in URI (URI = expression entered in address field of webbrowser)), like "Country=Germany&City=Aachen"
// and set the ConfigValues
int DecodeGETParameterAndSetConfigValues(String GETParameter)
{
   
   int posFirstCharToSearch = 1;
   int count = 0;
   
   // while a "&" is in the expression, after a start position to search
   while (GETParameter.indexOf('&', posFirstCharToSearch) > -1)
   {
     int posOfSeparatorChar = GETParameter.indexOf('&', posFirstCharToSearch);  // position of & after start position
     int posOfValueChar = GETParameter.indexOf('=', posFirstCharToSearch);      // position of = after start position
  
     ConfigValue[count] = GETParameter.substring(posOfValueChar + 1, posOfSeparatorChar);  // extract everything between = and & and enter it in the ConfigValue
      
     posFirstCharToSearch = posOfSeparatorChar + 1;  // shift the start position to search after the &-char 
     count++;
   }

   // no more & chars found
   
   int posOfValueChar = GETParameter.indexOf('=', posFirstCharToSearch);       // search for =
   
   ConfigValue[count] = GETParameter.substring(posOfValueChar + 1, GETParameter.length());  // extract everything between = and end of string
   count++;

   return count;  // number of values found in GET parameter
}








// +++++++++++++++++++ Start of Application Code +++++++++++++++++++

// include RGB LED library
#include <RGBLED.h>

// construct the object BreadboardRGBLED of class type RGBLED
RGBLED BreadboardRGBLED(PIN_LED_RED, PIN_LED_GREEN, PIN_LED_BLUE);


// include TCP Client library
#include <TCPClient.h>

// construct the object attTCPClient of class TCPClient
TCPClient attTCPClient = TCPClient();


// include MQTT Client library
#include <MQTTClient.h>

// construct the object attMQTTClient of class MQTTClient
MQTTClient attMQTTClient = MQTTClient();




#define NETWORKSTATE_LOGGED       1
#define NETWORKSTATE_NOTLOGGED    0

byte RouterNetworkDeviceState; 


#define NODESTATE_RESET                    3    // S3 = connection + subscription established after user reset
#define NODESTATE_PING                     2    // S2 = connection still open, successful ping
#define NODESTATE_OK                       1    // S1 = connection + subscription established
#define NODESTATE_NOTSPECIFIEDERROR        0    // E0 = unspecified error
#define NODESTATE_NOTCPCONNECTION          -1   // E1 = connection not established because of no tcp connection to test broker
#define NODESTATE_NOMQTTCONNECTION         -2   // E2 = connection not established because broker rejected MQTT connection attempt
#define NODESTATE_NOMQTTSUBSCRIPTION       -3   // E3 = broker rejected subscription attempt
#define NODESTATE_NOPING                   -4   // E4 = ping was not successful, connection lost

byte MQTTClient_Connected;
byte CounterForMQTT;





void setup() 
{
  attMQTTClient.myTCPClient = attTCPClient;    // tell the MQTT object the TCP object because MQTT library is using the TCP/IP library too
  
  pinMode(PIN_LED_BOARD, OUTPUT);
  pinMode(PIN_ACTOR, OUTPUT);

  
  preferences.begin("MyJourney", false);   // see https://github.com/espressif/arduino-esp32/blob/master/libraries/Preferences/examples/StartCounter/StartCounter.ino 

  // put the ESP32 in both modes
  WiFi_SetBothModesNetworkStationAndAccessPoint();

  // Switch the RGB LED to yellow (with a public function of our RGBLED object)
  BreadboardRGBLED.SwitchRGBLED(LED_YELLOW); 
 
  
  // Open serial communications and wait for port to open
  Serial.begin(19200);
  Serial.println("Hello ESP32!");


  // takeout 4 Strings out of the Non-volatile storage
  String strSSID = preferences.getString("SSID", "");
  String strPassword = preferences.getString("Password", "");
  String strDeviceID = preferences.getString("DeviceID", "");
  String strDeviceToken = preferences.getString("DeviceToken", "");

  // convert it to char*
  char* txtSSID = const_cast<char*>(strSSID.c_str());
  char* txtPassword = const_cast<char*>(strPassword.c_str());   // https://coderwall.com/p/zfmwsg/arduino-string-to-char 

  // try to connect to the LAN
  int success = WiFi_RouterNetworkConnect(txtSSID, txtPassword);
    
  if (success == 1)
  {
      BreadboardRGBLED.SwitchRGBLED(LED_YELLOW);    // switch the RGB LED color to yellow
      BreadboardRGBLED.setRGB(255, 255, 0);          // and set the default color to yellow for dimming
      RouterNetworkDeviceState = NETWORKSTATE_LOGGED;     // set the state    
  }
  else
  {
      BreadboardRGBLED.SwitchRGBLED(LED_RED);      // switch the RGB LED color to red
      RouterNetworkDeviceState = NETWORKSTATE_NOTLOGGED;  // set the state    
  }   


  // Start access point with SSID "ESP32_MyJourneyIoT"
  int success2 = WiFi_AccessPointStart("ESP32_MyJourneyIoT");

  // switch LED off
  SwitchActor(ACTOR_OFF);

  // initialize config values and set names
  InitializeConfigValues();
  ConfigName[0] = "LED";
  ConfigType[0] = 1;     // type textbox
  
  ConfigName[1] = "SSID";
  ConfigType[1] = 1;
  
  ConfigName[2] = "PASSWORD";
  ConfigType[2] = 1;
  
  ConfigName[3] = "DEVICE ID";
  ConfigType[3] = 1;         
  
  ConfigName[4] = "DEVICE TOKEN";
  ConfigType[4] = 1;
  
 
  
  // put the NVS stored values in RAM for the program
  ConfigValue[1] = strSSID;
  ConfigValue[2] = strPassword;
  ConfigValue[3] = strDeviceID;
  ConfigValue[4] = strDeviceToken;
 
  // start the webserver to listen for request of clients (in LAN or own ESP32 network)
  Webserver_Start();


  MQTTClient_Connected = false;
  CounterForMQTT = 0;


}



// check the ConfigValues and set ConfigStatus
// process the first ConfigValue to switch something
void ProcessAndValidateConfigValues(int countValues)
{
  if (countValues > 8) {countValues = 8;};

  // store the second to fifth value in non-volatile storage
  if (countValues > 4)
  {
    preferences.putString("SSID",        ConfigValue[1]);
    preferences.putString("Password",    ConfigValue[2]);
    preferences.putString("DeviceID",    ConfigValue[3]);
    preferences.putString("DeviceToken", ConfigValue[4]);
  }

  // in our application the first value must be "00" or "FF" (as text string)
  if ((ConfigValue[0].equals("00")) || (ConfigValue[0].equals("FF")))
  {
    ConfigStatus[0] = 1;    // Value is valid
  }
  else
  {
    ConfigStatus[0] = -1;   // Value is not valid
  }

  // first config value is used to switch LED ( = Actor)
  if (ConfigValue[0].equals("00"))   
  {
    SwitchActor(ACTOR_OFF);
  }
  
  if (ConfigValue[0].equals("FF"))
  {
    SwitchActor(ACTOR_ON);
  }

  
}





// Connect to AllThingsTalk MQTT Broker and switch RGB-LEDs
int ConnectToATT()
{   
    int NodeState;

    // closes TCP connection
    attTCPClient.Close();
       
    // TCP-connect to AllThingsTalk MQTT Server
    int TCP_Connect_Result = attTCPClient.Connect("api.allthingstalk.io", 1883);
    
    // Success?
    if (TCP_Connect_Result != 1)
    {
      // no = red LED
      BreadboardRGBLED.SwitchRGBLED(LED_RED);
      BreadboardRGBLED.setRGB(255, 0, 0);          // and set the default color to red for dimming 
      NodeState = NODESTATE_NOTCPCONNECTION;  
    }
    else
    {
      // yes = yellow LED
      BreadboardRGBLED.SwitchRGBLED(LED_YELLOW); 

      // take authentication strings for AllThingsTalk out of controller memory
      String strDeviceID = preferences.getString("DeviceID", "");
      String strDeviceToken = preferences.getString("DeviceToken", "");

      // send out MQTT CONNECT request with Device Data, Password is not needed by AllThingsTalk, we can set it to "abc"
      int MQTT_Connect_Result = attMQTTClient.Connect(strDeviceID, "maker:" + strDeviceToken, "abc");
      
      if (MQTT_Connect_Result == 1)
      {
        // connected
        BreadboardRGBLED.SwitchRGBLED(LED_GREEN);    // green LED
        BreadboardRGBLED.setRGB(0, 255, 0);          // and set the default color to green for dimming
        NodeState = NODESTATE_OK;

        Serial.println("MQTT connected.");        
      }
      else
      {
        // not connected
        BreadboardRGBLED.SwitchRGBLED(LED_RED);      // red LED
        BreadboardRGBLED.setRGB(255, 0, 0);          // and set the default color to red for dimming
        NodeState = NODESTATE_NOMQTTCONNECTION;  

        Serial.println("MQTT not connected.");
      }
   }

   return NodeState;
}



// Publish a MQTT message with SensorDate (here: 1 Byte, value 10..99) to AllThingsTalk MQTT Broker
// SensorName is the asset name for the AllThingsTalk user device
void SendSensorDateToATT(String SensorName, byte SensorDate)
{
    // Dummy for message payload in JSON format
    String MString = "{\"value\": 00.0}";
  
    int Slength = MString.length();

    // copy the bytes of the Dummy string in the payload buffer
    for (int j = 0; j < Slength; j++)
    {
        attMQTTClient.PayloadBuffer[j] = MString[j];
    }

    // encode SensorDate Byte (10..99) in ASCII format inside the payload JSON expression
    attMQTTClient.PayloadBuffer[10] = (SensorDate / 10) + 48;
    attMQTTClient.PayloadBuffer[11] = (SensorDate % 10) + 48;

    // Device ID for AllThingsTalk out of controller memory
    String strDeviceID = preferences.getString("DeviceID", "");

    // generate the appropriate Topic for AllThingsTalk MQTT Broker with Device data
    String ATT_SensorTopic = "device/" + strDeviceID + "/asset/" + SensorName + "/state";

    // Publish the message with topic and payload
    attMQTTClient.Publish(ATT_SensorTopic, attMQTTClient.PayloadBuffer, Slength);

}



int CurrentSensorDate = 10;

void loop() {
   
    String GETParameter = Webserver_GetRequestGETParameter();   // look for client request
  
    if (GETParameter.length() > 0)        // we got a request, client connection stays open
    {
      if (GETParameter.length() > 1)      // request contains some GET parameter
      {
          
          int countValues = DecodeGETParameterAndSetConfigValues(GETParameter);     // decode the GET parameter and set ConfigValues
  
          if (WebRequestHostAddress == "192.168.4.1")                               // the user entered this address in browser, with GET parameter values for configuration
          {
    
                ProcessAndValidateConfigValues(5);                                  // check and process 5 ConfigValues, switch the LED, store SSID, Password, DeviceID and DeviceToken in non-volatile storage
            
                String strSSID = preferences.getString("SSID", "");                 // takeout SSID and Password out of non-volatile storage  
                String strPassword = preferences.getString("Password", "");              
              
                char* txtSSID = const_cast<char*>(strSSID.c_str());                 // convert to char*
                char* txtPassword = const_cast<char*>(strPassword.c_str());         // https://coderwall.com/p/zfmwsg/arduino-string-to-char
  
                int successDisconnect = WiFi_RouterNetworkDisconnect();             // disconnect from router network
  
                BreadboardRGBLED.SwitchRGBLED(LED_YELLOW); 
                
                int successConnect = WiFi_RouterNetworkConnect(txtSSID, txtPassword);   // then try to connect once new with new login-data
                  
                if (successConnect == 1)
                {
                    BreadboardRGBLED.SwitchRGBLED(LED_YELLOW);                      // switch the RGB LED color to yellow
                    BreadboardRGBLED.setRGB(255, 255, 0);                           // and set the default color to yellow for dimming
                    RouterNetworkDeviceState = NETWORKSTATE_LOGGED;                 // set the state

                    CounterForMQTT = 0;
                }
                else
                {
                    BreadboardRGBLED.SwitchRGBLED(LED_RED);                         // switch RGB LED to red 
                    RouterNetworkDeviceState = NETWORKSTATE_NOTLOGGED;              // set the state                 
                }   
                
                MQTTClient_Connected = false;
          }
          
      }

      String HTMLPage;
       
      if (WebRequestHostAddress == "192.168.4.1")                  //   the user entered this address in the browser, to get the configuration webpage               
      {
          HTMLPage = EncodeFormHTMLFromValues("ESP32 Webserver CONFIG", 5) + "<br>IP Address: " + WiFi_GetOwnIPAddressInRouterNetwork();       // build a new Configuration webpage with form and new ConfigValues entered in textboxes or represented by sliders
      }
      
      Webserver_SendHTMLPage(HTMLPage);    // send out the webpage to client = webbrowser and close client connection
      
  }


  if (RouterNetworkDeviceState == NETWORKSTATE_LOGGED)   // if we are logged in
  {
    BreadboardRGBLED.dimRGB();                           // process a dimming step

    CounterForMQTT++;
    
    if (CounterForMQTT > 40)                             // do this approx every 2 seconds
    {
      CounterForMQTT = 0;
      
      // not yet connected
      if (MQTTClient_Connected == false)
      {
        int NodeState = ConnectToATT();   // try to connect to AllThingsTalk MQTT Broker and switch RGB-LEDs
    
        if (NodeState > 0)   //  connection and subscription established successfully
        {
          MQTTClient_Connected = true;            // we can go in normal mode 
        }
    
      }
    
    
      // we successfully connected with MQTT broker and subscribed to topic and can go in normal mode
      if (MQTTClient_Connected == true)
      {
           
          // generate virtual sensor data (10..40)
          CurrentSensorDate++;
          if (CurrentSensorDate > 40) {CurrentSensorDate = 10;};
    
          // Send sensor data via MQTT to AllThingsTalk
          SendSensorDateToATT("temperature", CurrentSensorDate);      

          // read the voltage at pin 36, divide by 64
          int analog_value = analogRead(PIN_SENSOR) >> 6;

          // Send sensor data via MQTT to AllThingsTalk
          SendSensorDateToATT("light", analog_value); 
        
      }

    }
    
  }
  
  delay(50);          // wait 50 ms
}
